// INCLUDES /////////////////////////////////////////////////////////////////////////////

// c++ includes: streams
#include <sstream>

// c++ maths
#include <cmath>
#define _USE_MATH_DEFINES
#include <math.h>

// class header
#include "playState.h"

// bell0bytes core
#include "app.h"
#include "depesche.h"
#include "coreComponent.h"
#include "window.h"

// bell0bytes file system
#include "fileSystemComponent.h"
#include "folders.h"

// bell0bytes input
#include "gameCommands.h"
#include "inputComponent.h"
#include "inputHandler.h"

// bell0bytes graphics
#include "graphicsComponent.h"
#include "graphicsComponent2D.h"
#include "d2d.h"
#include "graphicsComponentWrite.h"
#include "sprites.h"

// bell0bytes audio
#include "audioComponent.h"

// bell0bytes mathematics
#include "geometry.h"
#include "trigonometry.h"

// bell0bytes physics
#include "kinematics.h"

// CLASS METHODS ////////////////////////////////////////////////////////////////////////
namespace game
{
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// Constructor and Destructor ////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	PlayState::PlayState(core::DirectXApp& app, const std::wstring& name) : GameState(app, name)
	{ }
	PlayState::~PlayState()
	{ }
	PlayState& PlayState::createInstance(core::DirectXApp& app, const std::wstring& stateName)
	{
		static PlayState instance(app, stateName);
		return instance;
	}
	Table::Table(const float leftX, const float leftY, const float rightX, const float rightY, const float frictionS, const float frictionK, const graphics::GraphicsComponent2D& gc) : leftX(leftX), leftY(leftY), rightX(rightX), rightY(rightY), frictionCoeffS(frictionS), frictionCoeffK(frictionK), gc(gc)
	{
		width = rightX - leftX;
		height = rightY - leftY;
		centre.x = width * 0.5f;
		centre.y = height * 0.5f;
	}
	Ball::Ball(const float x, const float y, const float angle, const graphics::GraphicsComponent2D& gc) : gc(gc), isRolling(false), angle(mathematics::Trigonometry::degToRad(angle)), initialVelocity(0.0f)
	{
		position = new mathematics::Vector2F(x, y);
		velocity = new mathematics::Vector2F(0.0f, 0.0f);
		acceleration = new mathematics::Vector2F(0.0f, 0.0f);
	}
	Ball::Ball(const float x, const float y, const float angle, const float friction, const graphics::GraphicsComponent2D& gc) : gc(gc), isRolling(false), angle(mathematics::Trigonometry::degToRad(angle)), initialVelocity(0.0f)
	{
		position = new mathematics::Vector2F(x, y);
		velocity = new mathematics::Vector2F(0.0f, 0.0f);
		acceleration = new mathematics::Vector2F(-friction*gravity, -friction*gravity);
	}
	Ball::~Ball()
	{
		if(position)
			delete position;
		if(velocity)
			delete velocity;
		if(acceleration)
			delete acceleration;
	}
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// Ball and Table ////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	void Table::draw() const
	{
		// draw the border
		gc.drawRectangle(D2D1::Point2F(leftX, leftY), D2D1::Point2F(rightX, rightY), NULL, 5);

		// draw centre segment
		gc.drawRectangle(D2D1::Point2F(leftX + centre.x, leftY), D2D1::Point2F(leftX + centre.x, rightY), NULL, 2.5f);

		// draw centre circle
		gc.drawEllipse(leftX + centre.x, leftY + centre.y, width * 0.1f, width * 0.1f, NULL, 2.5f);

		// draw centre point
		gc.drawEllipse(leftX + centre.x, leftY + centre.y, 1, 1, NULL, 5);

		// draw Strafraum
		gc.drawRectangle(D2D1::Point2F(leftX, centre.y + leftY - 221.76f), D2D1::Point2F(leftX + 242, centre.y + leftY + 221.76f), NULL, 2.5f);
		gc.drawRectangle(D2D1::Point2F(leftX+width-242, centre.y + leftY - 221.76f), D2D1::Point2F(leftX + width, centre.y + leftY + 221.76f), NULL, 2.5f);

		// draw Torraum
		gc.drawRectangle(D2D1::Point2F(leftX, centre.y + leftY - 100.76f), D2D1::Point2F(leftX + 80.6f, centre.y + leftY + 100.76f), NULL, 2.5f);
		gc.drawRectangle(D2D1::Point2F(leftX + width - 80.66f, centre.y + leftY - 100.76f), D2D1::Point2F(leftX + width, centre.y + leftY + 100.76f), NULL, 2.5f);
	}

	void Table::decreaseFriction()
	{
		frictionCoeffK -= 0.1f;
	}

	void Table::increaseFriction()
	{
		frictionCoeffK += 0.1f;
	}

	void Ball::draw(const double farSeer, const double dt) const
	{
		// get oracled position
		float oracleX = position->x + velocity->x * (float)dt * (float)farSeer;
		float oracleY = position->y + velocity->y * (float)dt * (float)farSeer;

		// draw the ball
		gc.fillEllipse(oracleX, oracleY, radius, radius, 0.5f);
		gc.drawEllipse(oracleX, oracleY, radius, radius, NULL, 2.5f);
	}

	void Ball::kickOff(const float frictionS, const float frictionK)
	{
		if (isRolling)
			return;

		if (initialVelocity < frictionS)
			return;
		
		// compute starting velocity
		velocity->x = initialVelocity * std::cosf(angle);
		velocity->y = initialVelocity * std::sinf(angle);

		// set acceleration based on the friction of the table
		acceleration->x = frictionK * gravity;
		acceleration->y = frictionK * gravity;

		// adjust sign of the acceleration
		acceleration->x = std::copysignf(acceleration->x, -velocity->x);
		acceleration->y = std::copysignf(acceleration->y, -velocity->y);
				
		isRolling = true;
	}

	void Ball::collision()
	{
		// compute new starting velocity
		velocity->x = initialVelocity * std::cosf(angle);
		velocity->y = initialVelocity * std::sinf(angle);

		// adjust sign of the acceleration
		acceleration->x = std::copysignf(acceleration->x, -velocity->x);
		acceleration->y = std::copysignf(acceleration->y, -velocity->y);
	}

	void Ball::increaseStartingVelocity()
	{
		initialVelocity += 10.0f;
	}

	void Ball::decreaseStartingVelocity()
	{
		initialVelocity -= 10.0f;
		if (initialVelocity < 0.0f)
			initialVelocity = 0.0f;
	}

	void Ball::decreaseStartingAngle()
	{
		if (!isRolling)
		{
			angle += mathematics::Trigonometry::degToRad(5);
			if (angle > (float)(2 * M_PI))
				angle = 0;
		}
	}

	void Ball::increaseStartingAngle()
	{
		if (!isRolling)
		{
			angle -= mathematics::Trigonometry::degToRad(5);
			if (angle < 0)
				angle = (float)(2 * M_PI);
		}
	}

	void Ball::update(const double deltaTime, const float frictionK)
	{
		if (isRolling)
		{
			physics::Kinematics::semiImplicitEuler(*position, *velocity, *acceleration, deltaTime);

			// adjust sign of the acceleration
			acceleration->x = std::copysignf(acceleration->x, -velocity->x);
			acceleration->y = std::copysignf(acceleration->y, -velocity->y);

			if (velocity->getSquareLength() < 0.1f)
				reset(frictionK);
		}
	}
	
	void Ball::reset(const float frictionK)
	{
		angle = std::atanf(velocity->y / velocity->x);
		if (velocity->x < 0.0f)
			angle += (float)M_PI;

		velocity->x = 0.0f;
		velocity->y = 0.0f;

		acceleration->x = frictionK * gravity;
		acceleration->y = frictionK * gravity;

		isRolling = false;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// Initialization //////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::initialize()
	{
		// catch errors
		util::Expected<void> result;
		
		// allow only keyboard input
		dxApp.getInputComponent().getInputHandler().activeKeyboard = true;
		dxApp.getInputComponent().getInputHandler().activeMouse = false;

		// create text formats
		result = dxApp.getGraphicsComponent().getWriteComponent().createTextFormat(L"Lucida Handwriting", 16.0f, frictionFormat);
		if (!result.isValid())
			return result;

		// create the table
		table = new Table(80, 90, 1920-80, 1080-90, 0.5f, 0.5f, dxApp.getGraphicsComponent().get2DComponent());

		// create the ball
		ball = new Ball(dxApp.getGraphicsComponent().getVirtualWidth()*0.5f, dxApp.getGraphicsComponent().getVirtualHeight()*0.5f, 45, table->frictionCoeffK, dxApp.getGraphicsComponent().get2DComponent());

		// create text layouts
		std::wostringstream info;
		info.precision(2);
		info << "Position: ( " << ball->position->x << " , " << ball->position->y << " )" << " --- ";
		info << "Velocity: " << ball->velocity->getLength() << " --> ( " << ball->velocity->x << " , " << ball->velocity->y << " )" << " --- ";
		info << "Angle: " << mathematics::Trigonometry::radToDeg(ball->angle) << " --- " << "Gravity: " << ball->gravity << std::endl;
		info << "Friction (Static , Kinetic): ( " << table->frictionCoeffS << " , " << table->frictionCoeffK << " )" << " --- ";
		info << "Acceleration: " << ball->acceleration->getLength() << " -- > (" << ball->acceleration->x << " , " << ball->acceleration->y << " )" << std::endl;
		result = dxApp.getGraphicsComponent().getWriteComponent().createTextLayoutFromWStringStream(info, frictionFormat.Get(), (float)dxApp.getGraphicsComponent().getCurrentWidth(), 100, frictionLayout);
		if (!result.wasSuccessful())
			return result;

		// notify the main application class that the game is running
		isPaused = false;

		firstCreation = false;
		
		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////  Pause and Resume ///////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::pause()
	{
		isPaused = true; 
		return { };
	}
	
	util::Expected<void> PlayState::resume()
	{
		isPaused = false;
		return { };
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// User Input //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::handleInput(std::unordered_map<input::GameCommands, input::GameCommand&>& activeKeyMap)
	{
		// act on user input
		for (auto x : activeKeyMap)
		{
			switch (x.first)
			{

			case input::GameCommands::MoveLeft:
				ball->decreaseStartingAngle();
				break;

			case input::GameCommands::MoveRight:
				ball->increaseStartingAngle();
				break;

			case input::GameCommands::MoveUp:
				ball->increaseStartingVelocity();
				break;

			case input::GameCommands::MoveDown:
				ball->decreaseStartingVelocity();
				break;

			case input::GameCommands::Select:
				ball->kickOff(table->frictionCoeffS, table->frictionCoeffK);
				break;

			case input::GameCommands::RotateLeft:
				table->increaseFriction();
				break;

			case input::GameCommands::RotateRight:
				table->decreaseFriction();
				break;

			case input::GameCommands::Back:
				dxApp.quitGame();
				break;

			default:
				break;
			}
		}
		return { };	
	}

	util::Expected<void> PlayState::onMessage(const core::Depesche& depesche)
	{
		if (depesche.type == core::DepescheTypes::ActiveKeyMap)
		{
			input::InputHandler* ih = (input::InputHandler*)depesche.sender;

			if (!isPaused)
				if (!ih->isListening())
					return handleInput(ih->activeKeyMap);
		}
		
		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////// Update /////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::update(const double deltaTime)
	{
		if (isPaused)
			return { };

		// update the ball and table
		ball->update(deltaTime, table->frictionCoeffK);

		// check boundaries
		if (ball->position->x < table->leftX + ball->radius)
		{
			ball->position->x = table->leftX + ball->radius;
			ball->initialVelocity = ball->velocity->getLength();
			ball->angle = (float)(M_PI - ball->angle);
			ball->collision();
		}
		if (ball->position->x > table->rightX - ball->radius)
		{
			ball->position->x = table->rightX - ball->radius;
			ball->initialVelocity = ball->velocity->getLength();
			ball->angle = (float)(M_PI - ball->angle);
			ball->collision();
		}
		if (ball->position->y < table->leftY + ball->radius)
		{
			ball->position->y = table->leftY + ball->radius;
			ball->initialVelocity = ball->velocity->getLength();
			ball->angle = (float)(2*M_PI - ball->angle);
			ball->collision();
		}
		if (ball->position->y > table->rightY - ball->radius)
		{
			ball->position->y = table->rightY - ball->radius;
			ball->initialVelocity = ball->velocity->getLength();
			ball->angle = (float)(2*M_PI - ball->angle);
			ball->collision();
		}

		// update text
		std::wostringstream info;
		info.precision(4);
		info << "Position: ( " << ball->position->x << " , " << ball->position->y << " )" << " --- ";
		if (ball->isRolling)
		{
			info << "Velocity: " << ball->velocity->getLength() << " --- ";
			float angle = std::atanf(ball->velocity->y / ball->velocity->x);
			if (ball->velocity->x < 0.0f)
				angle += (float)M_PI;
			info << "Angle: " << mathematics::Trigonometry::angleInStandardCoordinates(mathematics::Trigonometry::radToDeg(angle)) << " --- " << "Gravity: " << ball->gravity << std::endl;
		}
		else
		{
			info << "Velocity: " << ball->initialVelocity << " --- ";
			info << "Angle: " << mathematics::Trigonometry::angleInStandardCoordinates(mathematics::Trigonometry::radToDeg(ball->angle)) <<  " --- " << "Gravity: " << ball->gravity << std::endl;
		}
		info << "Friction (Static , Kinetic): ( " << table->frictionCoeffS << " , " << table->frictionCoeffK << " )" << " --- ";
		info << "Acceleration: " << ball->acceleration->getLength() << " -- > ( " << ball->acceleration->x << " , " << ball->acceleration->y << " )" << std::endl;

		// create text layout
		if (!dxApp.getGraphicsComponent().getWriteComponent().createTextLayoutFromWStringStream(info, frictionFormat.Get(), (float)dxApp.getGraphicsComponent().getCurrentWidth(), 100, frictionLayout).wasSuccessful())
			return std::exception("Failed to create text layout!\n");

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// Render //////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::render(const double farSeer)
	{
		// render the table
		table->draw();

		// render the ball
		ball->draw(farSeer, dxApp.getPhysicsDeltaTime());

		// draw direction
		float oracledX = ball->position->x + ball->velocity->x * (float)farSeer * (float)dxApp.getPhysicsDeltaTime();
		float oracledY = ball->position->y + ball->velocity->y * (float)farSeer * (float)dxApp.getPhysicsDeltaTime();
		if (ball->isRolling)
		{
			float angle = std::atanf(ball->velocity->y / ball->velocity->x);
			if (ball->velocity->x < 0.0f)
				angle += (float)M_PI;
			dxApp.getGraphicsComponent().get2DComponent().drawLine(D2D1::Point2F(oracledX, oracledY), D2D1::Point2F(oracledX + std::cosf(angle) * 15, oracledY + std::sinf(angle) * 15), 2.5f);
		}
		else
			dxApp.getGraphicsComponent().get2DComponent().drawLine(D2D1::Point2F(oracledX, oracledY), D2D1::Point2F(oracledX + std::cosf(ball->angle) * 15, oracledY + std::sinf(ball->angle) * 15), 2.5f);
		
		// print text
		dxApp.getGraphicsComponent().getWriteComponent().printCenteredText(frictionLayout.Get(), -225, 560);

		// print FPS information
		dxApp.getGraphicsComponent().getWriteComponent().printFPS();

		// return success
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////// Shutdown ////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> PlayState::shutdown()
	{
		isPaused = true;

		// delete projectile
		if (table)
			delete table;
		if (ball)
			delete ball;
	
		// return success
		return { };
	}
}