#pragma once

/****************************************************************************************
* Author:	Gilles Bellot
* Date:		14/06/2018 - Lenningen - Luxembourg
*
* Desc:		main state of the running game
*
* Hist:
****************************************************************************************/

// INCLUDES /////////////////////////////////////////////////////////////////////////////

// c++ includes
#include <iostream>
#include <array>

// Microsoft
#include <wrl.h>

// direct write
#include <dwrite_3.h>

// bell0bytes core
#include "states.h"
#include "depesche.h"

// bell0bytes graphics
#include "graphicsComponent3D.h"
#include "graphicsComponent2D.h"

// bell0bytes mathematics
#include "vectors.h"

// bell0bytes physics
#include "kinematics.h"

// DEFINITIONS //////////////////////////////////////////////////////////////////////////
namespace UI
{
	class HeadsUpDisplayState;
}

namespace graphics
{
	class Sprite;
}

namespace core
{
	class DirectXApp;
	class DepescheSender;
	class DepescheDestination;
}

namespace audio
{
	struct SoundEvent;
};

namespace core
{
	class DirectXApp;
}

namespace audio
{
	struct SoundEvent;
	struct StreamEvent;
}

namespace physics
{
	class Projectile;
	class Kinematics;
}

namespace game
{
	class Table
	{
	private:
		// graphics component
		const graphics::GraphicsComponent2D& gc;				// address of the graphics component of the game engine

		// structure to define a virtual kicker table
		float leftX, leftY, rightX, rightY;						// the borders of the table (rectangle)
		float offsetX, offsetY;									// the offsets (top, bottom, left, right)
		float frictionCoeffS, frictionCoeffK;					// the static and kinetic friction coefficients
		float width, height;									// the height and the width of the table
		mathematics::Vector2F centre;							// the centre of the table

	public:
		// constructor
		Table(const float leftX, const float leftY, const float rightX, const float rightY, const float frictionS, const float frictionK, const graphics::GraphicsComponent2D& gc);

		// change friction
		void decreaseFriction();
		void increaseFriction();

		// draw
		void draw() const;

		friend class PlayState;
	};

	class Ball
	{
	private:
		// graphics component
		const graphics::GraphicsComponent2D& gc;					// address of the graphics component of the game engine

		// structure to define a ball
		const float radius = 11.0f;									// the radius of the ball
		mathematics::Vector2F* position;							// the position of the ball
		mathematics::Vector2F* velocity;							// the velocity of the ball
		mathematics::Vector2F* acceleration;						// the acceleration of the ball

		// movement variables
		const float gravity = 90.0f;								// gravity of the game world
		float angle;												// direction of movement in radians
		float initialVelocity;										// desired initial velocity

		// status
		bool isRolling;												// true if and only if the ball is actually rolling around

	public:
		// constructor and destructor
		Ball(const float x, const float y, const float angle, const graphics::GraphicsComponent2D& gc);									// sets the position of the ball to (x,y), sets the velocity and acceleration to 0
		Ball(const float x, const float y, const float angle, const float friction, const graphics::GraphicsComponent2D& gc);			// sets the position of the ball to (x,y) and its acceleration to (ax, ay) - the velocity is set to 0
		~Ball();

		// movement	
		void update(const double deltaTime, const float frictionK);	// updates the position of the ball depending on its velocity and acecleration using semi-implicit Euler integration
		void kickOff(const float frictionS, const float frictionK);
		void collision();
		void increaseStartingVelocity();
		void decreaseStartingVelocity();
		void increaseStartingAngle();
		void decreaseStartingAngle();
		void reset(const float frictionK);

		// draw
		void draw(const double farSeer, const double dt) const;

		friend class PlayState;
	};

	class PlayState : public core::GameState
	{
	private:
		// the constructor is private to ensure that this will be a singleton
		PlayState(core::DirectXApp& app, const std::wstring& name);

		// text
		Microsoft::WRL::ComPtr<IDWriteTextFormat3> frictionFormat;
		Microsoft::WRL::ComPtr<IDWriteTextLayout4> frictionLayout;

		// table and ball
		Table* table;
		Ball* ball;

	public:
		virtual ~PlayState();

		// singleton: get instance
		static PlayState& createInstance(core::DirectXApp& app, const std::wstring& name);

		// initialization
		virtual util::Expected<void> initialize() override;
		virtual util::Expected<void> shutdown() override;

		// pause and resume
		virtual util::Expected<void> pause() override;
		virtual util::Expected<void> resume() override;

		// user input
		virtual util::Expected<void> handleInput(std::unordered_map<input::GameCommands, input::GameCommand&>& activeKeyMap) override;
		virtual util::Expected<void> update(const double deltaTime) override;

		// render the scene
		virtual util::Expected<void> render(const double farSeer) override;

		// handle message
		virtual util::Expected<void> onMessage(const core::Depesche&) override;
	};
}